;
;	_setjmp.asm
;	
;	setjmp() and longjmp() for Windows
;	
;	
;	For the purpose of creating Foxpro/Win FLLs with Borland C++
;
;	Walter Nicholls
;   Cornerstone Software Limited
;
;	Revision History
;	----------------
;	 3 May 1993  WRN	Created
;	26 May 1994  WRN    Revised with BC++ 4.0 in mind
;	 
;	
;	Comments
;	------------------
;	Requires LARGE memory model (FAR code, FAR data)
;
;	This source code has been placed in the Public Domain.  It does not
;	contain any code proprietary to Borland, Microsoft or anyone else.
;
;	You may distribute and modify this source in any way you wish, however
;	we ask that:
;		1)	You do not make any charge for this software
;       2)  You acknowledge the original author in any derived source
;
;   To assemble this module use the following command line
;
;       tasm -ml _setjmp       
;

ideal
model large


; This structure is 18 bytes long, 2 shorter than Borland's. This is
; because I am not saving the flags word. 
; I have done this because the Foxpro LCK code (compiled with MSC) 
; allocates only 18 bytes for a jmp_buf, so if I tried to write 20
; bytes of data into it ... well, crasho.
;
struc jmp_buf           
	j_sp	dw ?
	j_ss	dw ?
; j_flag dw ?  under Borland C++ (DOS)
	j_cs	dw ?
	j_ip	dw ?
	j_bp	dw ?
	j_di	dw ?
	j_es	dw ?
	j_si	dw ?
	j_ds	dw ?
ends jmp_buf

;
;	void far cdecl longjmp(jmp_buf __jmpb, int __retval);
;	int  far cdecl setjmp(jmp_buf __jmpb);
;

codeseg




;
;	Note: this setjmp & longjmp do NOT touch the FPU!!!
;	
;	This may be a mistake .. but I doubt it matters for FLLs
;	
;


; *************************************************
;		setjmp() - saves CPU state
; *************************************************
;   
;	Saves all register variables, seg registers and flags
;   Does not bother saving ax/bx/cx/dx
;	
;	Setjmp is called with a sequence such as
;	
;			push	seg    jmp_buf
;			push    offset jmp_buf
;			call far _setjmp
;			add     sp, 4
;			
;	So the stack will appear on entry to setjmp() as
;	
;			Seg jmp_buf
;			offset jmp_buf
;			cs
;	sp -->	ip
;			
;	
;

proc _setjmp far

	push bp
	mov  bp, sp

; Offsets into stack frame at this point:
;
STK_OLDBP	equ	(bp+0)
STK_IP		equ	(bp+2)
STK_CS		equ	(bp+4)
STK_JMPBUFO	equ	(bp+6)
STK_JMPBUFS	equ	(bp+8)

	mov		cx, ds		; save caller's data segment in cx
	
	mov		ds, [STK_JMPBUFS]
	mov		bx, [STK_JMPBUFO]
	
	mov		[(jmp_buf bx).j_ds], cx
	
	mov		ax, [STK_CS]
	mov		[(jmp_buf bx).j_cs], ax
	mov		ax, [STK_IP]
	mov		[(jmp_buf bx).j_ip], ax

	mov		ax, [STK_OLDBP]
	mov		[(jmp_buf bx).j_bp], ax	

	mov		[(jmp_buf bx).j_si], si	
	mov		[(jmp_buf bx).j_di], di	
	mov		[(jmp_buf bx).j_es], es	
	mov		[(jmp_buf bx).j_ss], ss
	
	add		bp, 10		; get stack pointer as it was before call to setjmp()
						; this INCLUDES before pushing parameters!!
	mov		[(jmp_buf bx).j_sp], bp

	mov		ds, cx		; restore caller's ds	

	xor		ax, ax		; return value THIS way is always 0
	pop		bp			; and bp
	ret					; and return
	
endp _setjmp	





proc  _longjmp far
;	On entry to this function the stack will look like:
;
;			retval
;			Seg jmp_buf
;			offset jmp_buf
;			caller's cs
;	sp -->	caller's ip
;			
; longjmp() seriously mucks with the stack, so eschewing usual
; push bp / mov bp,sp sequence here in favour of something a
; little unconventional!

	add		sp, 4			; forget caller's CS:IP - we aren't returning
	pop		bx				; ds:bx = pointer to jmp_buf
	pop		ds
	pop		ax				; return value
	
; if ax is 0, set it to 1 - clever way of doing it that
; doesn't require a jump (and corresponding performance
; hit from stalling CPU pipeline, not that it matters)
;	
    cmp     ax, 1           ; generates carry if ax=0
    adc     ax, 0           ; if was 0, now it's 1

; restore old stack pointer, and set up stack frame again
	mov		ss, [(jmp_buf bx).j_ss]
	mov		sp, [(jmp_buf bx).j_sp]
	
	push 	ds				; put address of jmp_buf back on the stack
	push	bx				; as it was on entering original _setjmp
	
	push	[(jmp_buf bx).j_cs]	; put execution address back on stack
	push	[(jmp_buf bx).j_ip]
		
; restore all the saved registers
	mov		bp, [(jmp_buf bx).j_bp]
	mov		di, [(jmp_buf bx).j_di]
	mov		si, [(jmp_buf bx).j_si]
	mov		es, [(jmp_buf bx).j_es]
	
	push	[(jmp_buf bx).j_ds]
	pop		ds
	
	ret				; returns to setjmp's, not longjmp's, caller
endp _longjmp


public _setjmp, _longjmp

end
